package agent_test

import (
	"encoding/json"
	"testing"

	"github.com/aquasecurity/starboard/pkg/apis/aquasecurity/v1alpha1"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v15/internal/module/starboard_vulnerability/agent"
	report "gitlab.com/gitlab-org/security-products/analyzers/report/v3"
)

func TestSupportedScanners(t *testing.T) {
	assert.True(t, agent.IsSupportedScanner("Trivy"), "supports Trivy")
	assert.True(t, agent.IsSupportedScanner("trivy"), "is case-insenstive")
	assert.False(t, agent.IsSupportedScanner("Aqua"), "doesn't support unknown scanners")
}

func TestConvert(t *testing.T) {
	rawReport := []byte(`
{
	"kind": "VulnerabilityReport",
	"apiVersion": "aquasecurity.github.io/v1alpha1",
	"metadata": {
		"name": "deployment-nginx-nginx",
		"namespace": "default",
		"uid": "08e86776-d51e-4c81-906d-17254ca34c6a",
		"resourceVersion": "1743",
		"generation": 1,
		"creationTimestamp": "2021-07-02T13:42:12Z",
		"labels": {
			"starboard.container.name": "nginx",
			"starboard.resource.kind": "Deployment",
			"starboard.resource.name": "nginx-deployment",
			"starboard.resource.namespace": "default"
		},
		"ownerReferences": [
			{
				"apiVersion": "apps/v1",
				"kind": "Deployment",
				"name": "nginx",
				"uid": "73ba1385-a99a-45a1-8727-96b210fe20fa",
				"controller": true,
				"blockOwnerDeletion": false
			}
		],
		"managedFields": [
			{
				"manager": "starboard",
				"operation": "Update",
				"apiVersion": "aquasecurity.github.io/v1alpha1",
				"time": "2021-07-02T13:42:12Z",
				"fieldsType": "FieldsV1",
				"fieldsV1": {
					"f:metadata": {
						"f:labels": {
							".": {},
							"f:starboard.container.name": {},
							"f:starboard.resource.kind": {},
							"f:starboard.resource.name": {},
							"f:starboard.resource.namespace": {}
						},
						"f:ownerReferences": {
							".": {},
							"k:{\"uid\":\"73ba1385-a99a-45a1-8727-96b210fe20fa\"}": {
								".": {},
								"f:apiVersion": {},
								"f:blockOwnerDeletion": {},
								"f:controller": {},
								"f:kind": {},
								"f:name": {},
								"f:uid": {}
							}
						}
					},
					"f:report": {
						".": {},
						"f:artifact": {
							".": {},
							"f:repository": {},
							"f:tag": {}
						},
						"f:registry": {
							".": {},
							"f:server": {}
						},
						"f:scanner": {
							".": {},
							"f:name": {},
							"f:vendor": {},
							"f:version": {}
						},
						"f:summary": {
							".": {},
							"f:criticalCount": {},
							"f:highCount": {},
							"f:lowCount": {},
							"f:mediumCount": {},
							"f:unknownCount": {}
						},
						"f:vulnerabilities": {}
					}
				}
			}
		]
	},
	"report": {
		"updateTimestamp": null,
		"scanner": {
			"name": "Trivy",
			"vendor": "Aqua Security",
			"version": "0.16.0"
		},
		"registry": {
			"server": "index.docker.io"
		},
		"artifact": {
			"repository": "library/nginx",
			"tag": "1.16"
		},
		"summary": {
			"criticalCount": 0,
			"highCount": 0,
			"mediumCount": 2,
			"lowCount": 0,
			"noneCount": 0,
			"unknownCount": 0
		},
		"vulnerabilities": [
			{
				"vulnerabilityID": "CVE-2020-27350",
				"resource": "apt",
				"installedVersion": "1.8.2",
				"fixedVersion": "1.8.2.2",
				"severity": "MEDIUM",
				"title": "apt: integer overflows and underflows while parsing .deb packages",
				"primaryLink": "https://avd.aquasec.com/nvd/cve-2020-27350",
				"links": [],
				"score": 5.7
			},
			{
				"vulnerabilityID": "CVE-2020-3810",
				"resource": "apt",
				"installedVersion": "1.8.2",
				"fixedVersion": "1.8.2.1",
				"severity": "MEDIUM",
				"title": "",
				"primaryLink": "https://avd.aquasec.com/nvd/cve-2020-3810",
				"links": [],
				"score": 5.5
			}
		]
	}
}`)

	input := new(v1alpha1.VulnerabilityReport)
	err := json.Unmarshal(rawReport, input)
	require.NoError(t, err)

	actual, err := agent.Convert(input, 5000000000)
	require.NoError(t, err)

	expectedScanner := agent.TrivyScanner
	expectedScanner.Version = "0.16.0"

	expected := []*agent.Payload{
		{
			Vulnerability: &report.Vulnerability{
				Name:        "CVE-2020-27350",
				Message:     "CVE-2020-27350 in apt",
				Description: "apt: integer overflows and underflows while parsing .deb packages",
				Severity:    report.SeverityLevelMedium,
				Confidence:  report.ConfidenceLevelUnknown,
				Solution:    "Upgrade apt from 1.8.2 to 1.8.2.2",
				Location: report.Location{
					Dependency: &report.Dependency{
						Package: report.Package{Name: "apt"},
						Version: "1.8.2",
					},
					Image: "index.docker.io/library/nginx:1.16",
					KubernetesResource: &report.KubernetesResource{
						Namespace:     "default",
						Name:          "nginx-deployment",
						ContainerName: "nginx",
						Kind:          "Deployment",
						AgentID:       "5000000000",
					},
				},
				Identifiers: []report.Identifier{
					{
						Type:  report.IdentifierTypeCVE,
						Name:  "CVE-2020-27350",
						Value: "CVE-2020-27350",
						URL:   "https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2020-27350",
					},
				},
				Links: []report.Link{{URL: "https://avd.aquasec.com/nvd/cve-2020-27350"}},
			},
			Scanner: expectedScanner,
		},
		{
			Vulnerability: &report.Vulnerability{
				Name:        "CVE-2020-3810",
				Message:     "CVE-2020-3810 in apt",
				Description: "",
				Severity:    report.SeverityLevelMedium,
				Confidence:  report.ConfidenceLevelUnknown,
				Solution:    "Upgrade apt from 1.8.2 to 1.8.2.1",
				Location: report.Location{
					Dependency: &report.Dependency{
						Package: report.Package{Name: "apt"},
						Version: "1.8.2",
					},
					Image: "index.docker.io/library/nginx:1.16",
					KubernetesResource: &report.KubernetesResource{
						Namespace:     "default",
						Name:          "nginx-deployment",
						ContainerName: "nginx",
						Kind:          "Deployment",
						AgentID:       "5000000000",
					},
				},
				Identifiers: []report.Identifier{
					{
						Type:  report.IdentifierTypeCVE,
						Name:  "CVE-2020-3810",
						Value: "CVE-2020-3810",
						URL:   "https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2020-3810",
					},
				},
				Links: []report.Link{{URL: "https://avd.aquasec.com/nvd/cve-2020-3810"}},
			},
			Scanner: expectedScanner,
		},
	}

	assert.Equal(t, expected, actual, "parses report into correct structure")
}
